/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * Copyright (C) 1986, Sun Microsystems, Inc. 
 */

/*
 * Do the real work of the keyserver .
 * Store secret keys. Compute common keys,
 * and use them to decrypt and encrypt DES keys .
 * Cache the common keys, so the
 * expensive computation is avoided. 
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <des_crypt.h>

#include <gmp.h>      /* GNU multiprecission library */
#undef min
#include <mp.h>

extern char ROOTKEY[];
extern char KEYSTORE[];

static FILE *kf;

static char rootkey[KEYBYTES];
static int  haverootkey;

struct storedkey {
       short uid;
       char crypt[KEYBYTES];
     };

static MINT *MODULUS;

static char     *fetchsecretkey();
static keystatus pk_crypt();


/*
 * Set the modulus for all our Diffie-Hellman operations 
 */
setmodulus (char *modx)
{
  MODULUS = xtom (modx);
}


/*
 * Set the secretkey key for this uid 
 */
keystatus pk_setkey (short uid, keybuf skey)
{
  if (!storesecretkey(uid, skey))
     return (KEY_SYSTEMERR);
  return (KEY_SUCCESS);
}


/*
 * Encrypt the key using the public key associated with remote_name and the
 * secret key associated with uid. 
 */
keystatus pk_encrypt (short uid, char *remote_name, des_block *key)
{
  return pk_crypt (uid, remote_name, key, DES_ENCRYPT);
}


/*
 * Decrypt the key using the public key associated with remote_name and the
 * secret key associated with uid. 
 */
keystatus pk_decrypt (short uid, char *remote_name, des_block *key)
{
  return pk_crypt (uid, remote_name, key, DES_DECRYPT);
}


/*
 * Do the work of pk_encrypt && pk_decrypt 
 */
static keystatus pk_crypt (short uid, char *remote_name, des_block *key, int mode)
{
  char     *xsecret;
  char      xpublic[HEXKEYBYTES + 1];
  char      xsecret_hold[HEXKEYBYTES + 1];
  des_block deskey;
  int       err;
  MINT     *public;
  MINT     *secret;
  MINT     *common;
  char      zero[8];

  xsecret = fetchsecretkey (uid);
  if (xsecret == NULL)
  {
    memset (zero,0,sizeof(zero));
    xsecret = xsecret_hold;
    if (!getsecretkey("nobody", xsecret, zero) || xsecret[0] == 0)
       return (KEY_NOSECRET);
  }
  if (!getpublickey(remote_name,xpublic) && !getpublickey("nobody",xpublic))
     return (KEY_UNKNOWN);

  if (!readcache(xpublic, xsecret, &deskey))
  {
    public = xtom (xpublic);
    secret = xtom (xsecret);
    common = itom (0);
    pow (public, secret, MODULUS, common);
    extractdeskey (common, &deskey);
    writecache (xpublic, xsecret, &deskey);
    mfree (secret);
    mfree (public);
    mfree (common);
  }
  err = ecb_crypt (&deskey, key, sizeof(des_block), DES_HW | mode);
  if (DES_FAILED(err))
     return (KEY_SYSTEMERR);
  return (KEY_SUCCESS);
}


/*
 * Choose middle 64 bits of the common key to use as our des key, possibly
 * overwriting the lower order bits by setting parity. 
 */
static extractdeskey (MINT *ck, des_block *deskey)
{
  MINT *a;
  short r;
  int   i;
  short base = (1 << 8);
  char *k;

  a = itom (0);
  move (ck, a);
  for (i = 0; i < ((KEYSIZE - 64) / 2) / 8; i++)
     sdiv (a, base, a, &r);

  k = deskey->c;
  for (i = 0; i < 8; i++)
  {
    sdiv (a, base, a, &r);
    *k++ = r;
  }
  mfree (a);
  des_setparity (deskey);
}


/*
 * Key storage management
 */

struct secretkey_list {
       short  uid;
       char   secretkey[HEXKEYBYTES+1];
       struct secretkey_list *next;
     };

static struct secretkey_list *g_secretkeys;

/*
 * Fetch the secret key for this uid 
 */
static char * fetchsecretkey (short uid)
{
  struct secretkey_list *l;

  for (l = g_secretkeys; l; l = l->next)
  {
    if (l->uid == uid)
       return (l->secretkey);
  }
  return (NULL);
}

/*
 * Store the secretkey for this uid 
 */
int storesecretkey (short uid, keybuf key)
{
  struct secretkey_list  *new;
  struct secretkey_list **l;
  int    nitems = 0;

  for (l = &g_secretkeys; *l && (*l)->uid != uid; l = &(*l)->next)
     nitems++;

  if (*l == NULL)
  {
    new = malloc (sizeof(*new));
    if (new == NULL)
       return (0);
    new->uid  = uid;
    new->next = NULL;
    *l = new;
  }
  else
    new = *l;

  memcpy (new->secretkey,key,HEXKEYBYTES);
  new->secretkey[HEXKEYBYTES] = 0;
  seekitem (nitems);
  writeitem (uid, new->secretkey);
  return (1);
}


int hexdigit (int val)
{
  return ("0123456789abcdef"[val]);
}

void bin2hex (unsigned char *bin, unsigned char *hex, int size)
{
  int i;

  for (i = 0; i < size; i++)
  {
    *hex++ = hexdigit (*bin >> 4);
    *hex++ = hexdigit (*bin++ & 15);
  }
}

int hexval (char dig)
{
  if ('0' <= dig && dig <= '9')
     return (dig - '0');
  if ('a' <= dig && dig <= 'f')
     return (dig - 'a' + 10);
  if ('A' <= dig && dig <= 'F')
     return (dig - 'A' + 10);
  return (-1);
}

void hex2bin (unsigned char *hex, unsigned char *bin, int size)
{
  int i;

  for (i = 0; i < size; i++)
  {
    *bin    = hexval (*hex++) << 4;
    *bin++ |= hexval (*hex++);
  }
}


int openstore (void)
{
  kf = fopen (KEYSTORE, "r+");
  if (kf == NULL)
  {
    kf = fopen (KEYSTORE, "w+");
    if (kf == NULL)
       return (0);
  }
  setbuf (kf, NULL);
  return (1);
}


void readkeys (void)
{
  struct secretkey_list  *node;
  struct secretkey_list **l;
  char   secretkey [HEXKEYBYTES+1];
  int    uid;

  if (kf == NULL)
     return;

  l = &g_secretkeys;
  seekitem (0);

  while (readitem(&uid, secretkey))
  {
    node = malloc (sizeof(*node));
    if (node == NULL)
       return;

    node->uid = uid;
    memcpy (node->secretkey,secretkey,HEXKEYBYTES + 1);
    node->next = NULL;
    *l = node;
    l  = &node->next;
  }
}

void writekeys (void)
{
  struct secretkey_list *k;

  seekitem (0);
  for (k = g_secretkeys; k; k = k->next)
      writeitem (k->uid, k->secretkey);
}

void seekitem (int item)
{
  if (kf)
     fseek (kf, item * sizeof(struct storedkey), 0);
}

int writeitem (int uid, char *key)
{
  struct storedkey item;
  char   rootkey_tmp[KEYBYTES];
  int    reencrypt;
  
  if (kf == NULL)
     return (1);

  if (uid == 0)
  {
    writerootkey (key);
    hex2bin (key, rootkey_tmp, KEYBYTES);
    reencrypt = (haverootkey && memcmp(rootkey,rootkey_tmp,KEYBYTES) != 0);
    memcpy (rootkey,rootkey_tmp,KEYBYTES);
    haverootkey = 1;
    if (reencrypt)
    {
      writekeys();
      return (1);
    }
  }
  if (!haverootkey)
     return (1);

  item.uid = uid;
  hex2bin (key, item.crypt, KEYBYTES);
  ecb_crypt (rootkey, item.crypt, KEYBYTES, DES_ENCRYPT|DES_HW);
  return (fwrite(&item,sizeof(item),1,kf) >= 0);
}


int readitem (int *uidp, char *key)
{
  struct storedkey item;

  if (!haverootkey || kf == NULL)
     return (0);

  if (fread(&item, sizeof(item), 1, kf) != 1)
     return (0);

  *uidp = item.uid;
  ecb_crypt (rootkey, item.crypt, KEYBYTES, DES_DECRYPT|DES_HW);
  bin2hex (item.crypt, key, KEYBYTES);
  key[HEXKEYBYTES] = 0;
  return (1);
}
  
/*
 * Root users store their key in /etc/$ROOTKEY so
 * that they can auto reboot without having to be
 * around to type a password. Storing this in a file
 * is rather dubious: it should really be in the EEPROM
 * so it does not go over the net for diskless machines.
 */
void writerootkey (char *secret)
{
  char newline = '\n';
  int  fd = open (ROOTKEY, O_WRONLY|O_TRUNC|O_CREAT, 0);

  if (fd < 0)
     perror (ROOTKEY);
  else
  {
    if (write(fd, secret, strlen(secret)) < 0 ||
        write(fd, &newline, sizeof(newline)) < 0)
    {
      fprintf (stderr, "%s: ", ROOTKEY);
      perror ("write");
    }
    close (fd);
  }
}


/*
 * Exponential caching management
 */
struct cachekey_list {
       keybuf    secret;
       keybuf    public;
       des_block deskey;
       struct cachekey_list *next;
     };
static struct cachekey_list *g_cachedkeys;


/*
 * cache result of expensive multiple precision exponential operation 
 */
static writecache (char *pub, char *sec, des_block *deskey)
{
  struct cachekey_list *new = malloc (sizeof(struct cachekey_list));

  if (new == NULL)
     return;

  memcpy (new->public,pub,sizeof(keybuf));
  memcpy (new->secret,sec,sizeof(keybuf));
  new->deskey  = *deskey;
  new->next    = g_cachedkeys;
  g_cachedkeys = new;
}

/*
 * Try to find the common key in the cache 
 */
static int readcache (char *pub, char *sec, des_block *deskey)
{
  struct cachekey_list  *found;
  struct cachekey_list **l;

#define cachehit(pub,sec,list) (!memcmp(pub,(list)->public,sizeof(keybuf)) && \
                                !memcmp(sec,(list)->secret,sizeof(keybuf)))

  for (l = &g_cachedkeys; *l && !cachehit(pub,sec,*l); l = &(*l)->next)
      ;
  if (*l == NULL)
     return (0);

  found = *l;
  (*l) = (*l)->next;
  found->next  = g_cachedkeys;
  g_cachedkeys = found;
  *deskey = found->deskey;
  return (1);
}
